/* compile with: gcc -lGL -lglut -Wall -o texture texture.c */
#ifdef __APPLE__
#include <GLUT/glut.h>
#include <libpng/png.h> // http://ethan.tira-thompson.org/Mac_OS_X_Ports.html
#else
#include <png.h>
#include <GL/glut.h>
#endif
#include <cstdio>
#include <string>
#include <stdlib.h>
#include <math.h>
#include <time.h>
 
 
#define TEXTURE_LOAD_ERROR 0
static GLuint texture;
static int texWidth, texHeight;
static  png_byte *image_data;
 
using namespace std;

/** loadTexture
 *     loads a png file into an opengl texture object, using cstdio , libpng, and opengl.
 * 
 *     \param filename : the png file to be loaded
 *     \param width : width of png, to be updated as a side effect of this function
 *     \param height : height of png, to be updated as a side effect of this function
 * 
 *     \return GLuint : an opengl texture id.  Will be 0 if there is a major error,
 *                                     should be validated by the client of this function.
 * 
 */
GLuint loadTexture(const string filename, int &width, int &height) 
{
  //header for testing if it is a png
  png_byte header[8];
 
  //open file as binary
  FILE *fp = fopen(filename.c_str(), "rb");
  if (!fp) {
    return TEXTURE_LOAD_ERROR;
  }
 
  //read the header
  fread(header, 1, 8, fp);
 
  //test if png
  int is_png = !png_sig_cmp(header, 0, 8);
  if (!is_png) {
    fclose(fp);
    return TEXTURE_LOAD_ERROR;
  }
 
  //create png struct
  png_structp png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL,
					       NULL, NULL);
  if (!png_ptr) {
    fclose(fp);
    return (TEXTURE_LOAD_ERROR);
  }
 
  //create png info struct
  png_infop info_ptr = png_create_info_struct(png_ptr);
  if (!info_ptr) {
    png_destroy_read_struct(&png_ptr, (png_infopp) NULL, (png_infopp) NULL);
    fclose(fp);
    return (TEXTURE_LOAD_ERROR);
  }
 
  //create png info struct
  png_infop end_info = png_create_info_struct(png_ptr);
  if (!end_info) {
    png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp) NULL);
    fclose(fp);
    return (TEXTURE_LOAD_ERROR);
  }
 
  //png error stuff, not sure libpng man suggests this.
  if (setjmp(png_jmpbuf(png_ptr))) {
    png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
    fclose(fp);
    return (TEXTURE_LOAD_ERROR);
  }
 
  //init png reading
  png_init_io(png_ptr, fp);
 
  //let libpng know you already read the first 8 bytes
  png_set_sig_bytes(png_ptr, 8);
 
  // read all the info up to the image data
  png_read_info(png_ptr, info_ptr);
 
  //variables to pass to get info
  int bit_depth, color_type;
  png_uint_32 twidth, theight;
 
  // get info about png
  png_get_IHDR(png_ptr, info_ptr, &twidth, &theight, &bit_depth, &color_type,
	       NULL, NULL, NULL);
 
  //update width and height based on png info
  width = twidth;
  height = theight;
 
  // Update the png info struct.
  png_read_update_info(png_ptr, info_ptr);
 
  // Row size in bytes.
  int rowbytes = png_get_rowbytes(png_ptr, info_ptr);
 
  // Allocate the image_data as a big block, to be given to opengl
  image_data = new png_byte[rowbytes * height];
  if (!image_data) {
    //clean up memory and close stuff
    png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
    fclose(fp);
    return TEXTURE_LOAD_ERROR;
  }
 
  //row_pointers is for pointing to image_data for reading the png with libpng
  png_bytep *row_pointers = new png_bytep[height];
  if (!row_pointers) {
    //clean up memory and close stuff
    png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
    delete[] image_data;
    fclose(fp);
    return TEXTURE_LOAD_ERROR;
  }
  // set the individual row_pointers to point at the correct offsets of image_data
  for (int i = 0; i < height; ++i)
    row_pointers[height - 1 - i] = image_data + i * rowbytes;
 
  //read the png into image_data through row_pointers
  png_read_image(png_ptr, row_pointers);
 
  //Now generate the OpenGL texture object
  GLuint texture;
  glGenTextures(1, &texture);
  glBindTexture(GL_TEXTURE_2D, texture);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
 
  //clean up memory and close stuff
  png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
  delete[] row_pointers;
  fclose(fp);
 
  return texture;
}

void deleteData() 
{
  delete[] image_data;
}

void changeData()
{
  for (int j = 0; j < texHeight; j++) {
    for (int i = 0; i < texWidth; i++) {
      png_byte *color = image_data + 4 * (j * texWidth + i);
      float r,g,b, a;
      r = color[0];
      g = color[1];
      b = color[2];
      a = color[3];

      // C'est ici qu'il reste du travail...

      color[0] = r; 
      color[1] = g; 
      color[2] = b; 
      color[3] = a; 
    }
  }
}

void render()
{
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  /* fov, aspect, near, far */
  gluPerspective(60, 1, 1, 10);
  gluLookAt(0, 0, -2, /* eye */
	    0, 0, 2, /* center */
	    0, 1, 0); /* up */
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
 
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glPushAttrib(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glEnable(GL_TEXTURE_2D);
  changeData();
  glTexImage2D(GL_TEXTURE_2D,0, GL_RGBA, texWidth, texHeight, 0,
	       GL_RGBA, GL_UNSIGNED_BYTE, (GLvoid*) image_data);
 
  glBegin(GL_QUADS);
  glNormal3f(0.0, 0.0, 1.0);
  glTexCoord2d(-1, -1); glVertex3f(-1.0, -1.0, 0.0);
  glTexCoord2d(-1, 0); glVertex3f(-1.0, 1.0, 0.0);
  glTexCoord2d(0, 0); glVertex3f(1.0, 1.0, 0.0);
  glTexCoord2d(0, -1); glVertex3f(1.0, -1.0, 0.0);
  glEnd();
 
  glDisable(GL_TEXTURE_2D);
  glPopAttrib();
 
  glFlush();
 
  glutSwapBuffers();
}
 
void init()
{
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glShadeModel(GL_SMOOTH);
 
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glEnable(GL_DEPTH_TEST);
 
  glLightfv(GL_LIGHT0, GL_POSITION, (GLfloat[]){2.0, 2.0, 2.0, 0.0});
  glLightfv(GL_LIGHT0, GL_AMBIENT, (GLfloat[]){1.0, 1.0, 1.0, 0.0});
 
  texture = loadTexture("texture.png", texWidth, texHeight);
}
  
void resize(int w, int h)
{
  glViewport(0, 0, (GLsizei) w, (GLsizei) h);
}
 
void key(unsigned char key, int x, int y)
{
  if (key == 'q') exit(0);
}
 
int main(int argc, char *argv[])
{
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
  glutInitWindowSize(640, 480);
  glutCreateWindow("Texture demo - [q]uit");
 
  init();
  glutDisplayFunc(render);
  glutReshapeFunc(resize);
  glutKeyboardFunc(key);
 
  glutMainLoop();
 
  deleteData();
  return 0;
}

